{
Copyright (c) 1998-2001 Jonathan Revusky
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
   This product includes software developed by Jonathan Revusky
4. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.
5. The CopyrightNotice variable in the source code file SmartJGlobals.pas is not
   to be modified.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}

unit classread;

interface

// This unit provides the functionality of reading .class files and getting
// the necessary info out of them.

uses Classes, SysUtils;

const 
  CONSTANT_UTF8 = 1;
  CONSTANT_Integer = 3;
  CONSTANT_Float = 4;
  CONSTANT_Long = 5;
  CONSTANT_Double = 6;
  CONSTANT_Class = 7;
  CONSTANT_String = 8;
  CONSTANT_FieldRef = 9;
  CONSTANT_MethodRef = 10;
  CONSTANT_InterfaceMethodRef = 11;
  CONSTANT_NameAndType = 12;

  ACC_PUBLIC = 1;
  ACC_PRIVATE = 2;
  ACC_PROTECTED = 4;
  ACC_STATIC = 8;
  ACC_FINAL = 16;
  ACC_SYNCHRONIZED = 32;
  ACC_VOLATILE = 64;
  ACC_TRANSIENT = 128;
  ACC_NATIVE = 256;
  ACC_ABSTRACT = 1024;

  ACC_STATIC_PUBLIC = ACC_PUBLIC or ACC_STATIC;

// This is the magic number that starts a class file.
  MAGIC_NUMBER = -20268354;

  MAIN_SIG = '([Ljava/lang/String;)V';

type

  // Exceptions to be raised when errors are found.

  EFileFormatException = class(Exception);
  EBadMagicNumber = class(EFileFormatException);


  TAttribute = class
  private
    FNameIndex : Word;
    Flength : Integer;
    FData : Pointer;
    constructor Create(stream : TStream);
  public
    destructor Destroy; override;
  end;

  // represents an offset into the Constant_pool part of the classfile.

  TInfo = class(TList)
  private
    FaccessFlags, FNameIndex, FdescriptorIndex : Word;
    constructor Create(Stream : TStream);
    function getModifiers : String; virtual; abstract;
    function getAttribute(I : Word) : TAttribute;
  public
    property AccessFlags : Word read FaccessFlags;
    property NameIndex : Word read FNameIndex;
    property DescriptorIndex : Word read FdescriptorIndex;
    property Modifiers : String read getModifiers;
    property Attributes[I: Word] : TAttribute read getAttribute;
    destructor Destroy; override;
  end;

  TMethodInfo = class(TInfo)
  private
    function getModifiers : String; override;
  end;

  TFieldInfo = class(Tinfo)
  private
    function getModifiers : String; override;
  end;

  TFieldList = class(TList)
  private
    constructor Create(Stream : TStream);
    function getField(I : Word) : TInfo;
  public
    property Fields[I : Word] : TInfo read getField;
    destructor Destroy; override;
  end;


  TPoolItem = class
  private
    Ftype : byte;
    Fsize : Word;
    Fdata : Pointer;
    constructor Create(stream : TStream);
  public
    destructor Destroy; override;
  end;

  TConstantPool = class(Tlist)
  private
    function getPoolItem(i : Word) : TPoolItem;
    function getString(i : Word) : String;
    function getStringRef(i : Word) : String;
    property poolItems[i : Word] : TPoolItem read getPoolItem;
  public
    property stringRefs[i : Word] : String read getStringRef;
    property strings[i : Word] : String read getString;
    constructor Create(stream : TStream);
    destructor Destroy; override;
  end;

  TClassFile = class
  private
    FConstantPool : TConstantPool;
    FminorVersion, FmajorVersion : Word;
    FClassName, FSuperClass : String;
    FInterfaces {, FTypes} : TStringList;
    FFieldList, FMethodList : TFieldList;
    FAccessFlags : Word;
    FHasMain, alreadyLooked : boolean;
    function getMethodName(I: Word) : string;
    function getFieldName(I: Word) : string;
    function getMethodDescriptor(I: Word) : string;
    function getFieldDescriptor(I: Word) : string;
    function getFieldModifiers(I: Word) : string;
    function getMethodModifiers(I: Word) : string;
    function lookForMain : boolean;
  public
    constructor Create(stream : TStream);
    destructor Destroy; override;
    property ConstantPool : TConstantPool read FConstantPool;
    property FullClassName : String read FClassName;
    property SuperClass : String read FSuperClass;
    property FieldList : TFieldList read FFieldList;
    property MethodList : TFieldList read FMethodList;
    property MethodNames[I: Word] : String read getMethodName;
    property FieldNames[I: Word] : String read getFieldName;
    property FieldDescriptors[I : Word] : String read getFieldDescriptor;
    property MethodDescriptors[I : Word] : String read getMethodDescriptor;
    property FieldModifiers[I: Word] : string read getFieldModifiers;
    property MethodModifiers[I: Word] : string read getMethodModifiers;
    property hasMain : Boolean read lookForMain;
end;


implementation


  //throws an exception indicating premature end of file.
  procedure PrematureEOF; forward;

  // reads the next byte off the stream.
  function nextOne(Stream : TStream) : byte; forward;

  // reads the next two bytes off the stream and inverts them.
  function nextTwo(Stream : TStream) : Word; forward;

  // reads the next four bytes off the stream and inverts them.
  function nextFour(Stream : TStream) : Integer; forward;

type
  ppoolItem = ^TPoolItem;
  Pbytes = array[0..30000] of Byte;
  PPBytes = ^PBytes;

  constructor TpoolItem.Create(stream : TStream);
  var
    pbuf : PPBytes;
  begin
    Ftype := nextOne(stream);
    case FType of
      CONSTANT_UTF8 :
          FSize := nextTwo(stream);
      CONSTANT_Integer,
      CONSTANT_Float,
      CONSTANT_FieldRef,
      CONSTANT_MethodRef,
      CONSTANT_InterfaceMethodRef,
      CONSTANT_NameAndType :
        FSize := 4;
      CONSTANT_Long, CONSTANT_Double :
        FSize := 8;
      CONSTANT_Class, CONSTANT_String :
        FSize := 2;
      else
       Raise EFileFormatException.Create('Error reading constant pool');
    end;
    FData := allocMem(FSize+1);
    pbuf := FData;
    if stream.read(pbuf^, Fsize) < FSize then
      PrematureEOF;
  end;

  destructor TpoolItem.Destroy;
  begin
    FreeMem(FData);
    inherited Destroy;
  end;

  constructor TConstantPool.Create(stream : TStream);
  var
    total : Word;
    poolItem : TPoolItem;
  begin
    total := nextTwo(stream);
    Capacity := total;
    while count < total -1 do
    begin
      poolItem := TPoolItem.Create(stream);
      if (poolItem.FType = CONSTANT_LONG) or (poolItem.Ftype = CONSTANT_DOUBLE) then
        add(Nil); // A long or a double takes up 2 entries in the constant pool
        add(poolItem);
    end;
  end;

  destructor TConstantpool.Destroy;
  var
    I : integer;
  begin
    for I:=1 to Count do
    begin
      poolItems[I].Free;
    end;
    inherited Destroy;
  end;
  
  function TConstantpool.getStringRef(I : Word) : String;
  type
    Pindex = ^Word;
  var
    offsetItem : TPoolItem;
    offset : Word;
  begin
    offsetItem := poolItems[i];
    offset := swap(PIndex(offsetItem.FData)^);
    result := getString(offset);
  end;
  
  
  function TConstantpool.getString(I : Word) : String;
  var
    StringItem : TPoolItem;
    P : PChar;
  begin
    stringItem := poolItems[I];
    if stringItem.FType <> CONSTANT_UTF8 then
      raise EFileFormatException.Create('Not a valid UTF8 String entry');
    P := stringItem.FData;
    result := String(P);
  end;
  
  function TConstantPool.getPoolItem(i : Word) : TpoolItem;
  begin
    result := TPoolItem(Items[i-1]);
  end;
   
  
  constructor TAttribute.Create(stream : TStream);
  var
    bytesLeft: Integer;
  begin
    FNameIndex := nextTwo(stream);
    Flength := nextFour(stream);
    bytesLeft := stream.Size - stream.position;
    if FLength > bytesLeft then 
      FLength := bytesLeft; // Is this hack necessary?
    FData := allocMem(FLength);
    if Stream.Read(PPBytes(FData)^, FLength) < FLength then 
      PrematureEOF;
  end;
  
  destructor TAttribute.Destroy;
  begin
    if FLength >0 then 
      FreeMem(FData);
  end;
  
  constructor TFieldList.Create(Stream : TStream);
  var
   total : Word;
  begin
   total := nextTwo(stream);
   capacity := total;
   while count < total do
    begin
       add(TFieldInfo.Create(Stream));
    end;
  end;

  destructor TFieldList.Destroy;
  var
    I: Integer;
  begin
    for i:=1 to Count do 
      TFieldInfo(items[i-1]).Free;
  end;
  
  function TClassFile.getMethodName(I: Word) : String;
  var
    MethodInfo : TInfo;
  begin
    MethodInfo := MethodList.Fields[I];
    result := ConstantPool.Strings[MethodInfo.NameIndex];
  end;
  
  function TClassFile.getMethodDescriptor(I: Word) : String;
  var
    MethodInfo : TInfo;
  begin
    MethodInfo := MethodList.Fields[I];
    result := ConstantPool.Strings[MethodInfo.DescriptorIndex];
  end;
  
  
  function TClassFile.getFieldName(I: Word) : String;
  var
    FieldInfo : TInfo;
  begin
    FieldInfo := FieldList.Fields[I];
    result := ConstantPool.Strings[FieldInfo.NameIndex];
  end;
  
  function TClassFile.getFieldDescriptor(I: Word) : String;
  var
    FieldInfo : TInfo;
  begin
    FieldInfo := FieldList.Fields[I];
    result := ConstantPool.Strings[FieldInfo.DescriptorIndex];
  end;
  
  
  function TFieldList.getField(I: Word) : TInfo;
  begin
    result := TFieldInfo(Items[i-1]);
  end;
  
  
  constructor TInfo.Create(Stream : TStream);
  var
    total : Word;
  begin
    FAccessFlags := nextTwo(Stream);
    FNameIndex := nextTwo(Stream);
    FdescriptorIndex := nextTwo(Stream);
    total := nextTwo(Stream);
    capacity := total;
    while count < total do
      add(TAttribute.Create(stream));
  end;
  
  destructor TInfo.Destroy;
  var
    I: Integer;
  begin
    for i:=1 to Count do 
      TAttribute(items[i-1]).Free;
  end;
  
  function TInfo.getAttribute(I: Word) : TAttribute;
  begin
    result := TAttribute(Items[I-1]);
  end;
  
  function TFieldInfo.getModifiers : String;
  begin
    result := '';
    if (FAccessFlags and ACC_PUBLIC) <>0 then 
      result := 'public ';
    if (FAccessFlags and ACC_PROTECTED) <> 0 then 
      result := 'protected ';
    if (FAccessFlags and ACC_PRIVATE) <> 0 then 
      result := 'private ';
    if (FAccessFlags and ACC_STATIC) <> 0 then 
      result := result + 'static ';
    if (FAccessFlags and ACC_FINAL) <> 0 then 
      result := result + 'final ';
    if (FAccessFlags and ACC_VOLATILE) <> 0 then 
      result := result + 'volatile ';
    if (FAccessFlags and ACC_TRANSIENT) <> 0 
      then result := result + 'transient ';
  end;

  
  function TMethodInfo.getModifiers : String;
  begin
    result := '';
    if (FAccessFlags and ACC_ABSTRACT) <> 0 then 
      result := 'abstract ';
    if (FAccessFlags and ACC_SYNCHRONIZED) <> 0 then 
      result := result + 'synchronized ';
    if (FAccessFlags and ACC_NATIVE) <> 0 then 
      result := result + 'native ';
    if (FAccessFlags and ACC_PUBLIC) <>0 then 
      result := result + 'public ';
    if (FAccessFlags and ACC_PROTECTED) <> 0 then 
      result := result + 'protected ';
    if (FAccessFlags and ACC_PRIVATE) <> 0 then 
      result := result + 'private ';
    if (FAccessFlags and ACC_STATIC) <> 0 then 
      result := result + 'static ';
    if (FAccessFlags and ACC_FINAL) <> 0 then 
      result := result + 'final ';
  end;
  
  
  function TClassFile.getFieldModifiers(I: Word) : String;
  begin
    result := FieldList.Fields[I].getModifiers;
  end;
  
  
  function TClassFile.getMethodModifiers(I: Word) : String;
  begin
    result := MethodList.Fields[I].getModifiers;
  end;
  
  
  // reads in from an arbitrary stream and sets up the internal
  // classfile data.
  
  constructor TClassFile.Create(Stream : TStream);
  var
    I : Integer;
    thisOffset, superOffset, offset,
    ifcCount : Word;
  begin
    if nextFour(Stream) <> MAGIC_NUMBER then
      raise EBadMagicNumber.Create('Bad magic number.');
    FminorVersion := nextTwo(stream);
    FmajorVersion := nextTwo(stream);
    FConstantPool := TConstantPool.Create(stream);
    FAccessFlags := nextTwo(stream);
    thisOffset := nextTwo(stream);
    superOffset := nextTwo(stream);
    ifcCount := nextTwo(stream);
    FClassName := FConstantPool.stringRefs[thisOffset];
    FsuperClass := FConstantPool.stringRefs[superOffset];
    FInterfaces := TStringList.Create;
    for I := 1 to ifcCount do
      begin
        offset := nextTwo(stream);
        FInterfaces.add(FConstantPool.stringRefs[offset]);
      end;
    FFieldList := TFieldList.Create(stream);
    FMethodList := TFieldList.Create(stream);

    // won't read beyond here for now.
  end;

  destructor TClassFile.Destroy;
  begin
    FConstantPool.Free;
    FInterfaces.Free;
    FFieldList.Free;
    FMethodList.Free;
    inherited Destroy;
  end;
  
  function TClassFile.lookForMain : boolean;
  var
    I : Integer;
  begin
    if (not alreadyLooked) then
    begin
      for I:= 1 to MethodList.Count do
      begin
        if (MethodNames[I] = 'main') and (MethodDescriptors[I] = MAIN_SIG) then
        begin
          FhasMain := (MethodList.Fields[I].FAccessFlags and ACC_STATIC_PUBLIC) = ACC_STATIC_PUBLIC;
          break;
        end;
      end;
    end;
    alreadyLooked := true;
    result := FhasMain;
  end;
  
  procedure PrematureEOF;
  begin
    raise EFileFormatException.Create('premature end of file reached');
  end;
  
  function nextOne(Stream : TStream) : byte;
  begin
    if Stream.read(result, 1) < 1 then 
      PrematureEOF;
  end;
  
  function nextTwo(Stream : TStream) : Word;
  begin
    if Stream.read(result, 2) < 2 then 
      PrematureEOF;
    result := swap(result);
  end;
  
  function nextFour(Stream : TStream) : Integer;
  begin
    if Stream.read(result, 4) < 4 then 
      PrematureEOF;
    result := swap(result shr 16) + swap(result shl 16);
  end;
end.

