unit readjar;

interface

uses SysUtils, Classes, ZClasses;
  
type
  
  byteArray = array[0..MAXINT-1] of Byte;
  pbyteArray = ^byteArray;
  
  TZipEntry = class
  private
    Fversion : Word;
    FVersionNeeded : Word;
    FBitFlags : Word;
    FCompressionType : Word;
    FLastModifiedTime : Word;
    FLastModifiedDate : Word;
    FCrc : Integer;
    FCompressedSize : Integer;
    FUncompressedSize : Integer;
    FFilenameLength : Word;
    FextraFieldLength : Word;
    FFileCommentLength : Word;
    FDiskNumberstart : Word;
    FinternalFileAttr : Word;
    FexternalFileAttr : Integer;
    FLocalHeaderOffset : Integer;
    FFilename, FExtraField, FFileComment : string;
    constructor Create(var buf : Pchar ; boundary : PChar);
  end;
  
  TZipFile = class(TStringList)
  private
    FStream : TFileStream;
    FEntriesOffset, FEntriesSize : Integer;
    PEntryBuffer : PChar;
  public
    constructor Create(filename : TFileName);
    destructor Destroy; override;
    procedure addZipEntry(zipEntry : TZipEntry);

// returns a pointer to the contents of the zip entry.
    function getEntry(entryName : String) : TZipEntry;
    function getCompressedLength(entryName : String) : Integer;
    function getUncompressedLength(entryName : String) : Integer;

// returns the entry's file data as a stream.
    function getEntryDataByName(entryName : String) : TStream;
    function getEntryData(I : Integer) : TStream;
  end;
  

  EZipException = class(Exception);

  
  function getManifestMainClass(ZipFile : TZipFile) : String;

implementation

type
  PINT = ^Integer;
  PWord = ^Word;
  
const
  READ_BUF_SIZE = 64;
  END_SIG = $06054b50;
  END_HEADER_LENGTH = 22;
  ENTRY_SIG = $02014b50;
  LOCAL_SIG = $04034b50;
  MANIFEST_FILE = 'meta-inf/manifest.mf';
  


var
  readbuf : array[0..READ_BUF_SIZE-1] of Byte;
  
  
  destructor TZipFile.Destroy; 
  begin
    FStream.Free;
    inherited Destroy;
  end;
  
  constructor TZipFile.Create(filename : TFileName);
  var
    boundary, bytesRead, readPos, I : Integer;
    foundIt : Boolean;
    commentLength, numEntries : Word;
    P : Pchar;
  begin
    if not FileExists(filename) then
      raise EFOpenError.Create('cannot open ' + filename);

    FStream := TFileStream.Create(filename, fmOpenRead or fmShareDenyNone);
    boundary := fStream.Size - $FFFF;
    if boundary < 0 then boundary := 0; // the endheader cannot be more than 64K long.
  
    foundIt := False;
    readPos := FStream.Size-READ_BUF_SIZE;
  
    while (readPos > boundary) and (not Foundit) do
    begin
      FStream.Position := readPos;
      bytesRead := FStream.Read(readbuf, READ_BUF_SIZE);
      for I:=bytesRead-4 downto 0 do
      begin
        if Pint(@readBuf[I])^ = END_SIG then
        begin
          if readPos + I + 22 > FStream.Size then 
            Continue; //can't be end header
          FStream.Position := readPos + I + 20;
          FStream.Read(commentLength, 2);
          if FStream.Position + commentLength = FStream.Size then // Found it!
          begin
            FoundIt := True;
            Fstream.Position := readPos + I + 4;
            break;
          end;
        end;
      end;
      readPos := readPos - READ_BUF_SIZE + 4;
    end;

    if not FoundIt then
      raise EZipException.Create(filename + ' does not appear to be a valid zip file.');
  
    Fstream.read(I, sizeof(Integer));
    if I<>0 then 
      raise EZipException.Create('multiple disk archives not supported');
  
  // get values from the End header.
  
    FStream.read(numEntries, 2);
  
    FStream.Position := FStream.Position + 2;
    FStream.read(FEntriesSize, 4);
    FStream.read(FEntriesOffset, 4);
  
  
  // position stream at the offset of the central directory record.
    FStream.Position := FEntriesOffset;
    GetMem(PentryBuffer, FEntriesSize);
  
  // slurp the entries.
    FStream.read(PEntryBuffer^, FEntriesSize);
  
  // set up the table of entries.
  
    P := PEntryBuffer;
  
    for I:=1 to NumEntries do
      addZipEntry(TZipEntry.Create(P, PEntryBuffer + FEntriesSize));
  end;
  
  procedure TZipFile.addZipEntry(zipEntry : TZipEntry);
    begin
      addObject(zipEntry.fFilename, zipEntry);
    end;
  
  
  function TZipFile.getEntry(entryName : String) : TZipEntry;
  var
    I: Integer;
  begin
    I := indexOf(entryName);
    if I = -1 then
      raise EZipException.create(entryName + ' not found in archive.');
    result := TZipEntry(objects[I]);
  end;
  
  function TZipFile.getCompressedLength(entryName : String) : Integer;
  var
    zipentry : TZipEntry;
  begin
    zipentry := getEntry(entryName);
    result := zipEntry.FCompressedSize;
  end;
  
  function TZipFile.getUncompressedLength(entryName : String) : Integer;
  var
    zipentry : TZipEntry;
  begin
    zipentry := getEntry(entryName);
    result := zipEntry.FUnCompressedSize;
  end;

  function TZipFile.getEntryDataByName(EntryName : String) : TStream;
  var
    I : Integer;
  begin
    I := IndexOf(EntryName);
    if I >= 0 then 
      result := getEntryData(indexOf(EntryName))
    else
      result := Nil;
  end;
  
  function TZipFile.getEntryData(I : Integer) : TStream;
  const
    magic : array [0..1] of Byte = ($78, $9C);
  var
    zipentry : TZipEntry;
    compressedLength, uncompressedLength, sig : Integer;
    namelength, fieldlength : Word;
    MemStream : TStream;
    isCompressed : Boolean;
    CompressedStream : TDecompressionStream;
  begin
    zipEntry := TZipEntry(GetObject(I));
    compressedLength := zipentry.FCompressedSize;
    uncompressedLength := zipentry.FUncompressedSize;
    isCompressed := (compressedLength <> uncompressedLength);

    Fstream.Position := zipentry.FLocalHeaderOffset;
  
    Fstream.read(sig, 4);
    if sig <> LOCAL_SIG then
      raise EZipException.Create('invalid local header signature');
  
    FStream.Position := FStream.Position + 22;
  
    FStream.Read(namelength, 2);
    FStream.Read(fieldLength, 2);
    FStream.Position := FStream.Position + namelength + fieldLength;
  
  // The stream should be positioned to read the entry data now.
    if not isCompressed then
    begin
      Result := TMemoryStream.Create;
      Result.CopyFrom(FStream, UnCompressedLength);
      Result.Position := 0;
    end
    else begin
      MemStream := TMemoryStream.Create;
      MemStream.write(magic, 2);
      MemStream.CopyFrom(FStream, CompressedLength);
      MemStream.Position := 0;
      CompressedStream := TDecompressionStream.Create(MemStream);
      Result := TMemoryStream.Create;
      Result.CopyFrom(CompressedStream, UnCompressedLength);
      CompressedStream.Free;
      Result.Position := 0;
    end;
  end;
  
  constructor TZipEntry.Create(var buf : Pchar; boundary : PChar);
  var
    P: Pchar;
  begin
    if buf + 45 > boundary then
      raise EZipException.Create('file format problem in zipfile');
    if Pint(buf)^ <> ENTRY_SIG then
      raise EZipException.Create('invalid entry header signature');
    inc(buf, 4);
    FVersion := PWord(buf)^;
    inc(buf, 2);
    FVersionNeeded := PWord(buf)^;
    inc(buf, 2);
    FBitFlags := PWord(buf)^;
    inc(buf, 2);
    FCompressionType := PWord(buf)^;
    inc(buf, 2);
    FLastModifiedTime := PWord(buf)^;
    inc(buf, 2);
    FLastModifiedDate := PWord(buf)^;
    inc(buf, 2);
    FCrc := PInt(buf)^;
    inc(buf, 4);
    FCompressedSize := PInt(buf)^;
    inc(buf, 4);
    FUnCompressedSize := PInt(buf)^;
    inc(buf, 4);
    FFilenameLength := PWord(buf)^;
    inc(buf, 2);
    FextraFieldLength := PWord(buf)^;
    inc(buf, 2);
    FFileCommentLength := PWord(buf)^;
    inc(buf, 2);
    FDiskNumberstart := PWord(buf)^;
    inc(buf, 2);
    FinternalFileAttr := PWord(buf)^;
    inc(buf, 2);
    FexternalFileAttr := PInt(buf)^;
    inc(buf, 4);
    FLocalHeaderOffset := PInt(buf)^;
    inc(buf, 4);
  
    if buf + FFileNameLength + FextraFieldLength + FFileCommentLength > boundary then
      raise EZipException.Create('error in zip entry format');
  
    P := allocMem(1+FFilenameLength);
    move(buf^, P^, FFilenameLength);
    inc(buf, FFilenameLength);
    FFileName := String(P);
    FreeMem(P);
    P := allocMem(1+FextraFieldLength);
    move(buf^, P^, FextraFieldLength);
    inc(buf, FextraFieldLength);
    FExtraField := String(P);
    FreeMem(P);
    P := allocMem(1+FFileCommentLength);
    move(buf^, P^, FFileCommentLength);
    inc(buf, FFileCommentLength);
    FFileComment := String(P);
    FreeMem(P);
  end;


  function getManifestMainClass(ZipFile : TZipFile) : String;
  var
    Stream : TStream;
    I, ManifestIndex : Integer;
    PC : PChar;
    S : String;
  begin
    ManifestIndex := -1;
    result := '';
    for I:=0 to ZipFile.Count-1 do 
    begin
      if LowerCase(ZipFile[I]) = MANIFEST_FILE then
      begin
        ManifestIndex := I;
        Break;
      end;
    end;
    if ManifestIndex = -1  then Exit;
    Stream := ZipFile.getEntryData(ManifestIndex);
    PC := AllocMem(1+Stream.Size);
    Stream.read(PC^, Stream.Size);
    Stream.Free;
    S := PC;
    FreeMem(PC);
    I := Pos('Main-Class:', S);
    if I=0 then
      Exit;
    S := Copy(S, I, length(S));
    for I :=  1 to length(S) do
    begin
      if (S[I] = #32) or (S[I]=#9) or (S[I] = #10) or (S[I] = #13) then
        break;
    end;
    S := Trim(Copy(S, I, length(S)));
    result := S;
  end;
end.
