/*
Copyright (c) 2001 Jonathan Revusky
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
   This product includes software developed by Jonathan Revusky
4. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.smartj;

import java.lang.reflect.*;
import java.io.*;

/**
 * A ClassLoader that, when asked to load a class,
 * first looks for a newer .java file and recompiles it first, using
 * Mr. Jikes.
 */

public class SmartJClassLoader extends ClassLoader {

    File sourceBase;
    String compileClasspath;

    public SmartJClassLoader(String srcPath) {
        sourceBase = new File(srcPath);
        if (!sourceBase.isDirectory()) {
            throw new IllegalArgumentException("The source path specified: " 
                + srcPath
                + " is not a directory.");
        }
        String pathSep = System.getProperty("path.separator");
        compileClasspath = sourceBase.getAbsolutePath() + pathSep;
        compileClasspath += System.getProperty("java.class.path");
        compileClasspath += pathSep;
        String jHome = System.getProperty("java.home");
        File rtjar = new File(jHome, "lib/rt.jar");
        if (!rtjar.exists())
            rtjar = new File(jHome, "lib/classes.zip");
        if (rtjar.exists()) {
            compileClasspath += rtjar.getAbsolutePath();
        }
    }

    public synchronized Class loadClass(String name, boolean resolve)
	throws ClassNotFoundException
    {
	// First, check if the class has already been loaded
	Class c = findLoadedClass(name);
	if (c == null) {
            c = findClass(name);
	}
	if (resolve) {
	    resolveClass(c);
	}
	return c;
    }

    private boolean hasSource(String classname) {
        classname = classname.replace('.', File.separatorChar);
        if (classname.indexOf('$') != -1)
            classname = classname.substring(0, classname.indexOf('$'));
        File f = new File(sourceBase, classname + ".java");
        return f.exists();
    }

    public Class findClass(String name) throws ClassNotFoundException {
        // Now check if we have the source
        if (name.startsWith("java.") || name.startsWith("javax.swing."))
            return findSystemClass(name);
        if (hasSource(name)) try {
            recompile(name);
            byte[] b = loadClassData(name);
            return defineClass(name, b, 0, b.length);
        }
        catch (Exception e) {
        }
        return findSystemClass(name);
    }

    private byte[] loadClassData(String name) throws IOException {
        InputStream is = null;
        String clsFile = name.replace('.', '/') + ".class";
        File classFile = new File(sourceBase, clsFile);
        FileInputStream fis = new FileInputStream(classFile);
        byte[] bytes = new byte[(int) classFile.length()];
        fis.read(bytes);
        return bytes;
    }

    private void recompile(String name) {
        int dollInd = name.indexOf('$');
        if (dollInd >=0) {
            name = name.substring(0, dollInd);
        }
        name = name.replace('.', File.separatorChar);
        String src = name + ".java";
        String cls = name + ".class";
        File srcFile = new File(sourceBase, src);
        File clsFile = new File(sourceBase, cls);
        long srcModified = srcFile.lastModified();
        long clsModified = clsFile.lastModified();
        if (srcModified > clsModified) {
            System.err.println("recompiling " + srcFile.getAbsolutePath());
            Jikes.compile("-classpath " + compileClasspath + " +E " + srcFile.getAbsolutePath());
        }
    }

    public static void main(String[] args) {
        if (args.length <=1) {
            System.err.println("usage: java com.revusky.smartj.SmartJClassLoader <classname> <srcpath> args...");
        }
        else try {
            String[] args2 = new String[args.length-2];
            ClassLoader loader = new SmartJClassLoader(args[1]);
            Class cl = loader.loadClass(args[0]);
            Method m = cl.getMethod("main", new Class[] {args.getClass()});
            for (int i = 0; i<args2.length; i++) {
                args2[i] = args[i+2];
            }
            m.invoke(null, new Object[] {args2});
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }
}
